module TestData

using Test, DataFrames, Random, Statistics
const ≅ = isequal

@testset "constructors" begin
    df1 = DataFrame([[1, 2, missing, 4], ["one", "two", missing, "four"]], [:Ints, :Strs])
    df2 = DataFrame([[1, 2, missing, 4], ["one", "two", missing, "four"]])
    df3 = DataFrame([[1, 2, missing, 4]])
    df6 = DataFrame([[1, 2, missing, 4], [1, 2, missing, 4], ["one", "two", missing, "four"]],
                    [:A, :B, :C])
    df7 = DataFrame(x = [1, 2, missing, 4], y = ["one", "two", missing, "four"])
    @test size(df7) == (4, 2)
    @test df7[!, :x] ≅ [1, 2, missing, 4]

    #test_group("description functions")
    @test size(df6, 1) == 4
    @test size(df6, 2) == 3
    @test propertynames(df6) == [:A, :B, :C]
    @test propertynames(df2) == [:x1, :x2]
    @test propertynames(df7) == [:x, :y]

    #test_group("ref")
    @test df6[2, 3] == "two"
    @test ismissing(df6[3, 3])
    @test df6[2, :C] == "two"
    @test df6[!, :B] ≅ [1, 2, missing, 4]
    @test size(df6[:, [2,3]], 2) == 2
    @test size(df6[2,:], 1) == ncol(df6) # this is a DataFrameRow
    @test size(df6[2:2,:], 1) == 1
    @test size(df6[[1, 3], [1, 3]]) == (2, 2)
    @test size(df6[1:2, 1:2]) == (2, 2)
    @test size(first(df6,2)) == (2, 3)
    # lots more to do

    #test_group("assign")
    df6[!, 3] = ["un", "deux", "trois", "quatre"]
    @test df6[1, 3] == "un"
    df6[!, :B] = [4, 3, 2, 1]
    @test df6[1,2] == 4
    df6[!, :D] = [true, false, true, false]
    @test df6[1,4]
    select!(df6, Not(:D))
    @test propertynames(df6) == [:A, :B, :C]
    @test size(df6, 2) == 3

    #test_context("SubDataFrames")

    #test_group("constructors")
    # single index is rows
    sdf6a = view(df6, 1:1, :)
    sdf6b = view(df6, 2:3, :)
    sdf6c = view(df6, [true, false, true, false], :)
    @test size(sdf6a) == (1,3)
    sdf6d = view(df6, [1,3], [:B])
    @test size(sdf6d) == (2,1)
    sdf6e = view(df6, [0x01], :)
    @test size(sdf6e) == (1,3)
    sdf6f = view(df6, UInt64[1, 2], :)
    @test size(sdf6f) == (2,3)
    sdf6g = view(df6, [1,3], :B)
    sdf6h = view(df6, 1, :B)
    sdf6i = view(df6, 1, [:B])

    #test_group("ref")
    @test sdf6a[1,2] == 4

    #test_context("Within")
    #test_group("Associative")

    #test_group("DataFrame")
    Random.seed!(1)
    N = 20
    d1 = Vector{Union{Int64, Missing}}(rand(1:2, N))
    d2 = CategoricalArray(["A", "B", missing])[rand(1:3, N)]
    d3 = randn(N)
    d4 = randn(N)
    df7 = DataFrame([d1, d2, d3], [:d1, :d2, :d3])
    ref_d1 = unique(d1)

    #test_group("groupby")
    gd = groupby(df7, :d1)
    @test length(gd) == 2
    @test gd[1][:, :d2] ≅ d2[d1 .== ref_d1[1]]
    @test gd[2][:, :d2] ≅ d2[d1 .== ref_d1[2]]
    @test gd[1][:, :d3] == d3[d1 .== ref_d1[1]]
    @test gd[2][:, :d3] == d3[d1 .== ref_d1[2]]

    g1 = groupby(df7, [:d1, :d2])
    g2 = groupby(df7, [:d2, :d1])
    @test g1[1][:, :d3] == g2[1][:, :d3]

    res = 0.0
    for x in g1
        res += sum(x[:, :d1])
    end
    @test res == sum(df7[!, :d1])

    df10 = DataFrame([[1:4;], [2:5;],
                      ["a", "a", "a", "b" ], ["c", "d", "c", "d"]],
                     [:d1, :d2, :d3, :d4])

    gd = groupby(df10, [:d3], sort=true)
    ggd = groupby(gd[1], [:d3, :d4], sort=true) # make sure we can groupby SubDataFrames
    @test ggd[1][1, :d3] == "a"
    @test ggd[1][1, :d4] == "c"
    @test ggd[1][2, :d3] == "a"
    @test ggd[1][2, :d4] == "c"
    @test ggd[2][1, :d3] == "a"
    @test ggd[2][1, :d4] == "d"
end

@testset "completecases and dropmissing" begin
    df1 = DataFrame([Vector{Union{Int, Missing}}(1:4), Vector{Union{Int, Missing}}(1:4)])
    df2 = DataFrame([Union{Int, Missing}[1, 2, 3, 4], ["one", "two", missing, "four"]])

    @test df2[completecases(df2), :] == df2[[1, 2, 4], :]
    @test dropmissing(df2) == df2[[1, 2, 4], :]
    returned = dropmissing(df1)
    @test df1 == returned && df1 !== returned
    df2b = copy(df2)
    @test dropmissing!(df2b) === df2b
    @test df2b == df2[[1, 2, 4], :]
    df1b = copy(df1)
    @test dropmissing!(df1b) === df1b
    @test df1b == df1

    @test_throws ArgumentError completecases(DataFrame())
    @test_throws MethodError completecases(DataFrame(x=1), true)

    for cols in (:x2, "x2", [:x2], ["x2"], [:x1, :x2], ["x1", "x2"], 2, [2], 1:2,
                 [true, true], [false, true], :,
                 r"x2", r"x", Not(1), Not([1]), Not(Int[]), Not([]), Not(Symbol[]),
                 Not(1:0), Not([true, false]), Not(:x1), Not([:x1]))
        @test df2[completecases(df2, cols), :] == df2[[1, 2, 4], :]
        @test dropmissing(df2, cols) == df2[[1, 2, 4], :]
        returned = dropmissing(df1, cols)
        @test df1 == returned && df1 !== returned
        df2b = copy(df2)
        @test dropmissing!(df2b, cols) === df2b
        @test df2b == df2[[1, 2, 4], :]
        @test dropmissing(df2, cols) == df2b
        @test df2 != df2b
        df1b = copy(df1)
        @test dropmissing!(df1b, cols) === df1b
        @test df1b == df1
    end

    df = DataFrame(a=[1, missing, 3])
    sdf = view(df, :, :)
    @test dropmissing(sdf) == DataFrame(a=[1, 3])
    @test eltype(dropmissing(df, disallowmissing=false).a) == Union{Int, Missing}
    @test eltype(dropmissing(df, disallowmissing=true).a) == Int
    @test eltype(dropmissing(sdf, disallowmissing=false).a) == Union{Int, Missing}
    @test eltype(dropmissing(sdf, disallowmissing=true).a) == Int
    @test df ≅ DataFrame(a=[1, missing, 3]) # make sure we did not mutate df

    @test_throws ArgumentError dropmissing!(sdf)

    df2 = copy(df)
    @test dropmissing!(df, disallowmissing=true) === df
    @test dropmissing!(df2, disallowmissing=false) === df2
    @test eltype(df.a) == Int
    @test eltype(df2.a) == Union{Int, Missing}
    @test df.a == df2.a == [1, 3]

    a = [1,2]
    df = DataFrame(a=a, copycols=false)
    @test dropmissing!(df) === df
    @test a === df.a
    dfx = dropmissing(df)
    @test df == df
    @test dfx !== df
    @test dfx.a !== df.a
    @test a === df.a # we did not touch df

    b = Union{Int, Missing}[1, 2]
    df = DataFrame(b=b)
    @test eltype(dropmissing(df).b) == Int
    @test eltype(dropmissing!(df).b) == Int
end

@testset "merge" begin
    Random.seed!(1)
    df1 = DataFrame(a = shuffle!(Vector{Union{Int, Missing}}(1:10)),
                    b = rand(Union{Symbol, Missing}[:A,:B], 10),
                    v1 = Vector{Union{Float64, Missing}}(randn(10)))

    df2 = DataFrame(a = shuffle!(Vector{Union{Int, Missing}}(1:5)),
                    b2 = rand(Union{Symbol, Missing}[:A,:B,:C], 5),
                    v2 = Vector{Union{Float64, Missing}}(randn(5)))

    m1 = innerjoin(df1, df2, on = :a)
    @test m1[!, :a] == df1[!, :a][df1[!, :a] .<= 5] # preserves df1 order
    m2 = outerjoin(df1, df2, on = :a)
    @test m2[!, :a] == df1[!, :a] # preserves df1 order
    @test m2[!, :b] == df1[!, :b] # preserves df1 order
    @test m2[indexin(df1[!, :a], m2[!, :a]), :b] == df1[!, :b]
    @test m2[indexin(df2[!, :a], m2[!, :a]), :b2] == df2[!, :b2]
    @test m2[indexin(df1[!, :a], m2[!, :a]), :v1] == df1[!, :v1]
    @test m2[indexin(df2[!, :a], m2[!, :a]), :v2] == df2[!, :v2]
    @test all(ismissing, m2[map(x -> !in(x, df2[!, :a]), m2[!, :a]), :b2])
    @test all(ismissing, m2[map(x -> !in(x, df2[!, :a]), m2[!, :a]), :v2])

    df1 = DataFrame(a = Union{Int, Missing}[1, 2, 3],
                    b = Union{String, Missing}["America", "Europe", "Africa"])
    df2 = DataFrame(a = Union{Int, Missing}[1, 2, 4],
                    c = Union{String, Missing}["New World", "Old World", "New World"])

    m1 = innerjoin(df1, df2, on = :a)
    @test m1[!, :a] == [1, 2]

    m2 = leftjoin(df1, df2, on = :a)
    @test m2[!, :a] == [1, 2, 3]

    m3 = rightjoin(df1, df2, on = :a)
    @test m3[!, :a] == [1, 2, 4]

    m4 = outerjoin(df1, df2, on = :a)
    @test m4[!, :a] == [1, 2, 3, 4]

    # test with missings (issue #185)
    df1 = DataFrame()
    df1[!, :A] = ["a", "b", "a", missing]
    df1[!, :B] = Union{Int, Missing}[1, 2, 1, 3]

    df2 = DataFrame()
    df2[!, :A] = ["a", missing, "c"]
    df2[!, :C] = Union{Int, Missing}[1, 2, 4]

    m1 = innerjoin(df1, df2, on = :A)
    @test size(m1) == (3,3)
    @test m1[!, :A] ≅ ["a","a", missing]

    m2 = outerjoin(df1, df2, on = :A)
    @test size(m2) == (5,3)
    @test m2[!, :A] ≅ ["a", "b", "a", missing, "c"]
end

@testset "join tests" begin
    df1 = DataFrame(a = Union{Symbol, Missing}[:x,:y][[1,1,1,2,1,1]],
                    b = Union{Symbol, Missing}[:A,:B,:D][[1,1,2,2,1,3]],
                    v1 = 1:6)

    df2 = DataFrame(a = Union{Symbol, Missing}[:x,:y][[2,2,1,1,1,1]],
                    b = Union{Symbol, Missing}[:A,:B,:C][[1,2,1,2,3,1]],
                    v2 = 1:6)
    df2[1,:a] = missing

    m1 = innerjoin(df1, df2, on = [:a,:b])
    @test m1 == DataFrame(a=[:x,:x,:x,:x,:x,:y,:x,:x],
                          b=[:A,:A,:A,:A,:B,:B,:A,:A],
                          v1=[1,1,2,2,3,4,5,5],
                          v2=[3,6,3,6,4,2,3,6])
    m2 = outerjoin(df1, df2, on = [:a,:b])
    @test m2 ≅ DataFrame(a=[:x,:x,:x,:x,:x,:y,:x,:x,:x,missing,:x],
                         b=[:A,:A,:A,:A,:B,:B,:A,:A,:D,:A,:C],
                         v1=[1,1,2,2,3,4,5,5,6,missing,missing],
                         v2=[3,6,3,6,4,2,3,6,missing,1,5])

    Random.seed!(1)
    function spltdf(d)
        d[!, :x1] = map(x -> x[1], d[!, :a])
        d[!, :x2] = map(x -> x[2], d[!, :a])
        d[!, :x3] = map(x -> x[3], d[!, :a])
        d
    end
    df1 = DataFrame(a = ["abc", "abx", "axz", "def", "dfr"], v1 = randn(5))
    df2 = DataFrame(a = ["def", "abc","abx", "axz", "xyz"], v2 = randn(5))
    spltdf(df1)
    spltdf(df2)

    m1 = innerjoin(df1, df2, on = :a, makeunique=true)
    m2 = innerjoin(df1, df2, on = [:x1, :x2, :x3], makeunique=true)
    @test m1[!, :a] == m2[!, :a]
end

@testset "nonunique, nonunique, unique! with extra argument" begin
    df1 = DataFrame(a = Union{String, Missing}["a", "b", "a", "b", "a", "b"],
                    b = Vector{Union{Int, Missing}}(1:6),
                    c = Union{Int, Missing}[1:3;1:3])
    df = vcat(df1, df1)
    @test findall(nonunique(df)) == collect(7:12)
    @test findall(nonunique(df, :)) == collect(7:12)
    @test findall(nonunique(df, Colon())) == collect(7:12)
    @test findall(nonunique(df, :a)) == collect(3:12)
    @test findall(nonunique(df, "a")) == collect(3:12)
    @test findall(nonunique(df, [:a, :c])) == collect(7:12)
    @test findall(nonunique(df, ["a", "c"])) == collect(7:12)
    @test findall(nonunique(df, r"[ac]")) == collect(7:12)
    @test findall(nonunique(df, Not(2))) == collect(7:12)
    @test findall(nonunique(df, Not([2]))) == collect(7:12)
    @test findall(nonunique(df, Not(:b))) == collect(7:12)
    @test findall(nonunique(df, Not([:b]))) == collect(7:12)
    @test findall(nonunique(df, Not([false, true, false]))) == collect(7:12)
    @test findall(nonunique(df, [1, 3])) == collect(7:12)
    @test findall(nonunique(df, 1)) == collect(3:12)

    @test unique(df) == df1
    @test unique(df, :) == df1
    @test unique(df, Colon()) == df1
    @test unique(df, 2:3) == df1
    @test unique(df, 3) == df1[1:3,:]
    @test unique(df, [1, 3]) == df1
    @test unique(df, [:a, :c]) == df1
    @test unique(df, ["a", "c"]) == df1
    @test unique(df, r"[ac]") == df1
    @test unique(df, Not(2)) == df1
    @test unique(df, Not([2])) == df1
    @test unique(df, Not(:b)) == df1
    @test unique(df, Not([:b])) == df1
    @test unique(df, Not([false, true, false])) == df1
    @test unique(df, :a) == df1[1:2,:]
    @test unique(df, "a") == df1[1:2,:]
    @test_throws ArgumentError unique(DataFrame())
    @test_throws ArgumentError nonunique(DataFrame())

    @test unique(copy(df1), "a") == unique(copy(df1), :a) == unique(copy(df1), 1) ==
          df1[1:2, :]

    unique!(df, [1, 3])
    @test df == df1
    for cols in (r"[ac]", Not(:b), Not(2), Not([:b]), Not([2]), Not([false, true, false]))
        df = vcat(df1, df1)
        unique!(df, cols)
        @test df == df1
    end
end

@testset "filter() and filter!()" begin
    df = DataFrame(x = [3, 1, 2, 1], y = ["b", "c", "a", "b"])
    @test filter(r -> r[:x] > 1, df) == DataFrame(x = [3, 2], y = ["b", "a"])
    @test filter!(r -> r[:x] > 1, df) === df == DataFrame(x = [3, 2], y = ["b", "a"])

    df = DataFrame(x = [3, 1, 2, 1], y = ["b", "c", "a", "b"])
    @test filter(:x => x -> x > 1, df) == DataFrame(x = [3, 2], y = ["b", "a"])
    @test filter!(:x => x -> x > 1, df) === df == DataFrame(x = [3, 2], y = ["b", "a"])

    df = DataFrame(x = [3, 1, 2, 1], y = ["b", "c", "a", "b"])
    @test filter("x" => x -> x > 1, df) == DataFrame(x = [3, 2], y = ["b", "a"])
    @test filter!("x" => x -> x > 1, df) === df == DataFrame(x = [3, 2], y = ["b", "a"])

    df = DataFrame(x = [3, 1, 2, 1], y = ["b", "c", "a", "b"])
    @test filter(1 => x -> x > 1, df) == DataFrame(x = [3, 2], y = ["b", "a"])
    @test filter!(1 => x -> x > 1, df) === df == DataFrame(x = [3, 2], y = ["b", "a"])

    df = DataFrame(x = [3, 1, 2, 1], y = ["b", "c", "a", "b"])
    @test filter([:x] => x -> x > 1, df) == DataFrame(x = [3, 2], y = ["b", "a"])
    @test filter!([:x] => x -> x > 1, df) === df == DataFrame(x = [3, 2], y = ["b", "a"])

    df = DataFrame(x = [3, 1, 2, 1], y = ["b", "c", "a", "b"])
    @test filter(["x"] => x -> x > 1, df) == DataFrame(x = [3, 2], y = ["b", "a"])
    @test filter!(["x"] => x -> x > 1, df) === df == DataFrame(x = [3, 2], y = ["b", "a"])

    df = DataFrame(x = [3, 1, 2, 1], y = ["b", "c", "a", "b"])
    @test filter((:) => (r...) -> r[1] > 1, df) == DataFrame(x = [3, 2], y = ["b", "a"])
    @test filter!((:) => (r...) -> r[1] > 1, df) === df == DataFrame(x = [3, 2], y = ["b", "a"])

    df = DataFrame(x = [3, 1, 2, 1], y = ["b", "c", "a", "b"])
    @test filter([:x, :x] => ==, df) == df
    @test filter!([:x, :x] => ==, df) === df == DataFrame(x = [3, 1, 2, 1], y = ["b", "c", "a", "b"])

    df = DataFrame(x = [3, 1, 2, 1], y = ["b", "c", "a", "b"])
    @test filter([2, 2] => !=, df) == DataFrame(x=Int[], y=String[])
    @test filter!([2, 2] => !=, df) === df == DataFrame(x=Int[], y=String[])

    for sel in [r"x", [1,2], [:x1, :x2], ["x1", "x2"], :, Not(r"y")]
        df = DataFrame(x1 = [3, 1, 2, 1], x2 = ["b", "c", "aa", "bbb"])
        @test filter(sel => (a, b) -> a == length(b), df) ==
              DataFrame(x1=[1, 2], x2=["c", "aa"])
        @test filter!(sel => (a, b) -> a == length(b), df) === df ==
              DataFrame(x1=[1, 2], x2=["c", "aa"])
    end

    df = DataFrame(x = [3, 1, 2, 1, missing], y = ["b", "c", "a", "b", "c"])
    @test_throws TypeError filter(r -> r[:x] > 1, df)
    @test_throws TypeError filter!(r -> r[:x] > 1, df)
    @test_throws TypeError filter(:x => x -> x > 1, df)
    @test_throws TypeError filter("x" => x -> x > 1, df)
    @test_throws TypeError filter!(:x => x -> x > 1, df)
    @test_throws TypeError filter!("x" => x -> x > 1, df)
    @test_throws TypeError filter(1 => x -> x > 1, df)
    @test_throws TypeError filter!(1 => x -> x > 1, df)
    @test_throws TypeError filter([:x] => x -> x > 1, df)
    @test_throws TypeError filter(["x"] => x -> x > 1, df)
    @test_throws TypeError filter!([:x] => x -> x > 1, df)
    @test_throws TypeError filter!(["x"] => x -> x > 1, df)
    @test_throws TypeError filter((:) => (r...) -> r[1] > 1, df)
    @test_throws TypeError filter!((:) => (r...) -> r[1] > 1, df)
end

@testset "filter and filter! with AsTable" begin
    df = DataFrame(x = [3, 1, 2, 1], y = ["b", "c", "a", "b"])

    function testfun(x)
        @assert x isa NamedTuple
        @assert propertynames(x) == (:x,)
        return x.x > 1
    end

    @test filter(AsTable(:x) => testfun, df) == DataFrame(x=[3, 2], y=["b", "a"])
    filter!(AsTable(:x) => testfun, df)
    @test df == DataFrame(x=[3, 2], y=["b", "a"])

    df = DataFrame(x = [3, 1, 2, 1], y = ["b", "c", "a", "b"])

    @test filter(AsTable("x") => testfun, df) == DataFrame(x=[3, 2], y=["b", "a"])
    filter!(AsTable("x") => testfun, df)
    @test df == DataFrame(x=[3, 2], y=["b", "a"])

    @test_throws ArgumentError filter([] => () -> true, df)
    @test_throws ArgumentError filter(AsTable(r"z") => () -> true, df)
    @test_throws ArgumentError filter!([] => () -> true, df)
    @test_throws ArgumentError filter!(AsTable(r"z") => () -> true, df)
end

@testset "names with cols" begin
    df = DataFrame(a = 1, x1 = 2, x2 = 3, x3 = 4, x4 = 5)

    for v in [df, groupby(df, :a)]
        @test names(v, All()) == names(v, :) == names(v) == ["a", "x1", "x2", "x3", "x4"]
        @test names(v, Between(:x1, :x3)) == ["x1", "x2", "x3"]
        @test names(v, Not(:a)) == names(v, r"x") == ["x1", "x2", "x3", "x4"]
        @test names(v, :x1) == names(v, 2) == ["x1"]
    end

    for v in [view(df, :, [4,3,2,1]), groupby(view(df, :, [4,3,2,1]), 1), view(df, 1, [4,3,2,1])]
        @test names(v, All()) == names(v, :) == names(v) ==  ["x3", "x2", "x1", "a"]
        @test names(v, Between(:x2, :x1)) == ["x2", "x1"]
        @test names(v, Not(:a)) == names(v, r"x") == ["x3", "x2", "x1"]
        @test names(v, :x1) == names(v, 3) == ["x1"]
    end
end

end # module
